import unittest

from shrug_lang.shrug_token import Token, TokenType
from shrug_lang.token_parser import TokenParser


class TokenGenerator:
    @staticmethod
    def get_eol():
        return Token(TokenType.EOL)

    @staticmethod
    def get_id(name='a'):
        return Token(TokenType.ID, name)

    @staticmethod
    def get_number(value=123):
        return Token(TokenType.NUMBER, value)

    @staticmethod
    def get_bool(value=True):
        return Token(TokenType.BOOL, value)

    @staticmethod
    def get_string(value='abc'):
        return Token(TokenType.STRING, value)

    @staticmethod
    def get_shrug():
        return Token(TokenType.SHRUG)

    @staticmethod
    def get_invalid(value='invalid'):
        return Token(TokenType.INVALID, value)

    @staticmethod
    def get_indent(value=0):
        return Token(TokenType.INDENT, value)


class BaseTokenParserTestCase(unittest.TestCase):
    """Base class for token parser test cases"""
    def setUp(self):
        self.token_parser = TokenParser()

    def process_tokens(self, tokens):
        return list(filter(
            lambda x: x is not None,
            [self.token_parser.next_token(token) for token in tokens]))



import unittest

from shrug_lang.shrug_token import TokenType, Token
from shrug_lang.tokenizer import Tokenizer


class TestTokenizer(unittest.TestCase):
    """Line parser class should return correctly read tokens"""

    def test_empty_line(self):
        line = ''
        self.assertEqual([Token(TokenType.INDENT, 0), Token(TokenType.EOL)],
                          Tokenizer.parse_line(line))

        line = '     '
        self.assertEqual([Token(TokenType.INDENT, 5), Token(TokenType.EOL)],
                         Tokenizer.parse_line(line))

    def test_shrug(self):
        line = '¯\_(ツ)_/¯'
        single_shrug = [Token(TokenType.INDENT, 0),
                        Token(TokenType.SHRUG),
                        Token(TokenType.EOL)]
        self.assertEqual(single_shrug, Tokenizer.parse_line(line))

    def test_numbers(self):
        line = '2'
        single_number1 = [Token(TokenType.INDENT, 0),
                          Token(TokenType.NUMBER, 2),
                          Token(TokenType.EOL)]
        self.assertEqual(single_number1, Tokenizer.parse_line(line))

        line = '43435356'
        single_number2 = [Token(TokenType.INDENT, 0),
                          Token(TokenType.NUMBER, 43435356),
                          Token(TokenType.EOL)]
        self.assertEqual(single_number2, Tokenizer.parse_line(line))

        line = 'dfs3'
        invalid_id1 = [Token(TokenType.INDENT, 0),
                       Token(TokenType.INVALID, 'dfs3'),
                       Token(TokenType.EOL)]
        self.assertEqual(invalid_id1, Tokenizer.parse_line(line))

        line = 'abc%'
        invalid_id2 = [Token(TokenType.INDENT, 0),
                       Token(TokenType.INVALID, 'abc%'),
                       Token(TokenType.EOL)]
        self.assertEqual(invalid_id2, Tokenizer.parse_line(line))

    def test_bool(self):
        line = 'False'
        just_false = [Token(TokenType.INDENT, 0),
                      Token(TokenType.BOOL, False),
                      Token(TokenType.EOL)]
        self.assertEqual(just_false, Tokenizer.parse_line(line))

        line = 'True'
        just_false = [Token(TokenType.INDENT, 0),
                      Token(TokenType.BOOL, True),
                      Token(TokenType.EOL)]
        self.assertEqual(just_false, Tokenizer.parse_line(line))

        line = 'False True True False'
        just_false = [Token(TokenType.INDENT, 0),
                      Token(TokenType.BOOL, False),
                      Token(TokenType.BOOL, True),
                      Token(TokenType.BOOL, True),
                      Token(TokenType.BOOL, False),
                      Token(TokenType.EOL)]
        self.assertEqual(just_false, Tokenizer.parse_line(line))

        line = 'false true'
        just_false = [Token(TokenType.INDENT, 0),
                      Token(TokenType.ID, 'false'),
                      Token(TokenType.ID, 'true'),
                      Token(TokenType.EOL)]
        self.assertEqual(just_false, Tokenizer.parse_line(line))

    def test_ids(self):
        line = 'a'
        single_id1 = [Token(TokenType.INDENT, 0),
                      Token(TokenType.ID, 'a'),
                      Token(TokenType.EOL)]
        self.assertEqual(single_id1, Tokenizer.parse_line(line))

        line = 'variable'
        single_id2 = [Token(TokenType.INDENT, 0),
                      Token(TokenType.ID, 'variable'),
                      Token(TokenType.EOL)]
        self.assertEqual(single_id2, Tokenizer.parse_line(line))

    def test_strings(self):
        line = '""'
        empty_string = [Token(TokenType.INDENT, 0),
                        Token(TokenType.STRING, ''),
                        Token(TokenType.EOL)]
        self.assertEqual(empty_string, Tokenizer.parse_line(line))

        line = '"a"'
        short_string = [Token(TokenType.INDENT, 0),
                        Token(TokenType.STRING, 'a'),
                        Token(TokenType.EOL)]
        self.assertEqual(short_string, Tokenizer.parse_line(line))

        line = '"a bc def gh i jk lmn op q rs tuv wx u y z"'
        long_string = [
            Token(TokenType.INDENT, 0),
            Token(TokenType.STRING, 'a bc def gh i jk lmn op q rs tuv wx u y z'),
            Token(TokenType.EOL)
        ]
        self.assertEqual(long_string, Tokenizer.parse_line(line))

        line = '"fs\'d\' f" " s " " fd_ f"'
        weird_string = [Token(TokenType.INDENT, 0),
                        Token(TokenType.STRING, 'fs\'d\' f'),
                        Token(TokenType.STRING, ' s '),
                        Token(TokenType.STRING, ' fd_ f'),
                        Token(TokenType.EOL)]
        self.assertEqual(weird_string, Tokenizer.parse_line(line))

        line = '"a'
        invalid_string = [Token(TokenType.INDENT, 0),
                          Token(TokenType.INVALID, '"a'),
                          Token(TokenType.EOL)]
        self.assertEqual(invalid_string, Tokenizer.parse_line(line))

        line = '"¯\_(ツ)_/¯"'
        shrug_string = [Token(TokenType.INDENT, 0),
                        Token(TokenType.STRING, '¯\_(ツ)_/¯'),
                        Token(TokenType.EOL)]
        self.assertEqual(shrug_string, Tokenizer.parse_line(line))

    def test_complex(self):
        line = ('¯\_(ツ)_/¯ n ¯\_(ツ)_/¯ x ¯\_(ツ)_/¯ ¯\_(ツ)_/¯ ¯\_(ツ)_/¯'
                ' 4 var "mm" 434')
        tokens = [Token(TokenType.INDENT, 0),
                  Token(TokenType.SHRUG), Token(TokenType.ID, 'n'),
                  Token(TokenType.SHRUG), Token(TokenType.ID, 'x'),
                  Token(TokenType.SHRUG), Token(TokenType.SHRUG),
                  Token(TokenType.SHRUG), Token(TokenType.NUMBER, 4),
                  Token(TokenType.ID, 'var'), Token(TokenType.STRING, 'mm'),
                  Token(TokenType.NUMBER, 434), Token(TokenType.EOL)]
        self.assertEqual(tokens, Tokenizer.parse_line(line))

    def test_join_strings(self):
        # trivial
        self.assertEqual([], Tokenizer.join_strings([]))
        self.assertEqual([''], Tokenizer.join_strings(['']))
        # unfinished strings
        self.assertEqual(['"'], Tokenizer.join_strings(['"']))
        self.assertEqual(['"affsdf'], Tokenizer.join_strings(['"affsdf']))
        self.assertEqual(['"affsdf abc'],
                         Tokenizer.join_strings(['"affsdf', 'abc']))
        # proper strings
        self.assertEqual(['"abc"'], Tokenizer.join_strings(['"abc"']))
        self.assertEqual(['"abc "def ghi"'],
                         Tokenizer.join_strings(['"abc', '"def', 'ghi"']))
        self.assertEqual(
            ['abc', '3', '"abc "def ghi"', 'hello', "ppp"],
            Tokenizer.join_strings(
                ['abc', '3', '"abc', '"def', 'ghi"', 'hello', "ppp"]))
        self.assertEqual(
            ['"     "'],
            Tokenizer.join_strings(['"', '', '', '', '', '"']))


if __name__ == '__main__':
    unittest.main()



from typing import Union

from shrug_lang.errors import TokenError
from shrug_lang.operators import CompOp, MathOp
from shrug_lang.parser_state import ParserState, StateTransformer
from shrug_lang.shrug_token import Token, TokenType


class TokenParser:
    def __init__(self):
        self.operations = {
            ParserState.ADDITION: MathOp.add,
            ParserState.SUBTRACTION: MathOp.subtract,
            ParserState.MULTIPLICATION: MathOp.multiply,
            ParserState.DIVISION: MathOp.divide,
            ParserState.MODULUS: MathOp.modulus,
            ParserState.EQ: CompOp.eq,
            ParserState.NEQ: CompOp.neq,
            ParserState.GT: CompOp.gt,
            ParserState.GTE: CompOp.gte,
            ParserState.LT: CompOp.lt,
            ParserState.LTE: CompOp.lte,
        }
        self.state_transformer = StateTransformer()
        self.value_map = {}
        self.state = ParserState.EMPTY
        self.tokens = []
        self.assign_to = None
        self.current_value = None
        self.value1 = None
        self.value2 = None
        self.in_cond = False
        self.skip_line = False

        # Not reset after line
        self.indent = 0
        self.skip_indented = False  # skip lines that are indented +1

    def reset_state(self):
        self.state = ParserState.EMPTY
        self.tokens = []
        self.assign_to = None
        self.current_value = None
        self.value1 = None
        self.value2 = None
        self.in_cond = False
        self.skip_line = False

    def next_token(self, token: Token):
        try:
            if token.type == TokenType.EOL:
                try:
                    if self.state == ParserState.EMPTY:
                        if self.assign_to:
                            return self.get_value(self.assign_to)
                        return
                    if self.state == ParserState.END:
                        if self.assign_to:
                            self.set_value(self.assign_to, self.current_value)
                        elif self.in_cond:
                            if self.current_value:
                                self.indent += 1
                            else:
                                self.skip_indented = True
                        else:
                            return self.current_value
                        return
                finally:
                    self.reset_state()
                # Not in valid state to end line
                raise TokenError('Unexpected end-of-line')

            if token.type == TokenType.INVALID:
                raise TokenError(f'Invalid token {token.value}')

            if self.skip_line:
                pass

            elif token.type == TokenType.INDENT:
                if token.value == self.indent:
                    if self.skip_indented:
                        self.skip_indented = False
                elif token.value < self.indent:
                    self.indent = token.value
                elif token.value > self.indent and self.skip_indented:
                    self.skip_line = True
                else:
                    raise TokenError('Unexpected indentation')

            elif self.state == ParserState.EMPTY:
                if token.type == TokenType.ID:
                    self.assign_to = token.value
                elif (token.type == TokenType.NUMBER or
                      token.type == TokenType.STRING or
                      token.type == TokenType.BOOL):
                    self.current_value = token.value

            elif (self.state == ParserState.MATH or
                  self.state == ParserState.COMPARE):
                if token.type == TokenType.ID:
                    self.value1 = (self.get_value(token.value), token.value)
                elif (token.type == TokenType.NUMBER or
                      token.type == TokenType.STRING or
                      token.type == TokenType.BOOL):
                    self.value1 = (token.value,)

            elif self.state in self.operations:
                operation = self.operations[self.state]
                if token.type == TokenType.ID:
                    self.value2 = (self.get_value(token.value), token.value)
                    self.current_value = operation(self.value1, self.value2)
                elif (token.type == TokenType.NUMBER or
                      token.type == TokenType.STRING or
                      token.type == TokenType.BOOL):
                    self.value2 = (token.value,)
                    self.current_value = operation(self.value1, self.value2)

            if self.state == ParserState.COND:
                if token.type == TokenType.ID:
                    self.current_value = self.get_value(token.value)
                    if self.assign_to:
                        raise TokenError('Cannot assign conditional statement')
                elif (token.type == TokenType.NUMBER or
                      token.type == TokenType.STRING or
                      token.type == TokenType.BOOL):
                    self.current_value = token.value
                    if self.assign_to:
                        raise TokenError('Cannot assign conditional statement')
                self.in_cond = True

            self.next_state(token)

        except Exception as e:
            self.reset_state()
            raise e

    def next_state(self, token: Token):
        self.state = self.state_transformer.next_state(self.state, token)

    def get_value(self, key: str) -> Union[int, str, None]:
        return self.value_map.get(key)

    def set_value(self, key: str, value: Union[int, str, None]):
        self.value_map[key] = value



import unittest

from shrug_lang.operators import BaseOp


class TestBaseOperations(unittest.TestCase):
    def test_operation_check_undefined(self):
        self.assertTrue(BaseOp.check_for_undefined((1, 'a'), (2, 'b')))
        self.assertRaises(ValueError, BaseOp.check_for_undefined,
                          (None, None), (2, 'b'))
        self.assertRaises(ValueError, BaseOp.check_for_undefined,
                          (1, 'a'), (None, None))
        self.assertRaises(ValueError, BaseOp.check_for_undefined,
                          (None, None), (None, None))

    def test_check_non_numbers(self):
        self.assertTrue(BaseOp.check_non_numbers(1, 2))
        self.assertRaises(TypeError, BaseOp.check_non_numbers, 'a', 1)
        self.assertRaises(TypeError, BaseOp.check_non_numbers, 2, '1')
        self.assertRaises(TypeError, BaseOp.check_non_numbers, '2', '4')

    def test_check_matching_types(self):
        class A:
            pass

        class B(A):
            pass

        self.assertTrue(BaseOp.check_matching_type(A(), A()))
        self.assertTrue(BaseOp.check_matching_type(1, 2))
        self.assertTrue(BaseOp.check_matching_type('a', 'bc'))

        self.assertRaises(TypeError, BaseOp.check_matching_type, 'a', 1)
        self.assertRaises(TypeError, BaseOp.check_matching_type, 2, '1')
        self.assertRaises(TypeError, BaseOp.check_matching_type, A(), B())


if __name__ == '__main__':
    unittest.main()
